from .selfsup.VICREGLoss import VICREGLoss
from .selfsup.ConVIRTLoss import ConVIRTLoss
from .selfsup.InfoNCELoss import InfoNCELoss
from .selfsup.GLoRIALoss import GLoRIALoss, gloria_attention_fn, cosine_similarity
from .mvqa.LabelSmoothingCrossEntropyLoss import LabelSmoothingCrossEntropy
from torch.nn.modules.loss import *
