from vilmedic.models.mvqa.MVQA import MVQA

from vilmedic.models.rrs.SumRNN import SumRNN
from vilmedic.models.rrs.SumHugMulti import SumHugMulti
# from vilmedic.models.rrs.RRS_obsolete import RRS
from vilmedic.models.rrs.RRS import RRS
from vilmedic.models.rrs.RRS_SCST import RRS_SCST
from vilmedic.models.rrs.RRS_SCST import RRS_SCST
from vilmedic.models.rrs.SumHugMono_SCST import SumHugMono_SCST

from vilmedic.models.selfsup.clip.VAE import VAE
from vilmedic.models.selfsup.clip.DALLE import DALLE
from vilmedic.models.selfsup.clip.CLIP import CLIP

from vilmedic.models.rrg.RRG import RRG
from vilmedic.models.rrg.RRG_FORCE import RRG_FORCE
from vilmedic.models.rrg.RRG_MULTI import RRG_MULTI
from vilmedic.models.rrg.RRG_EEG import RRG_EEG
from vilmedic.models.rrg.RRG_PPO import RRG_PPO
from vilmedic.models.rrg.RRG_SCST import RRG_SCST
from vilmedic.models.rrg.RRG_2D_CT import RRG_2D_CT

from vilmedic.models.selfsup.conVIRT import ConVIRT
from vilmedic.models.selfsup.SimCLR import SimCLR
from vilmedic.models.selfsup.GLoRIA import GLoRIA
from vilmedic.models.selfsup.VICREG import VICREG
from vilmedic.models.selfsup.SwaV import SwaV
from vilmedic.models.selfsup.PCL import PCL
from vilmedic.models.selfsup.PCL2 import PCL2

